------------------------------------------------------------------------------
--	FILE:	 RandomisedBritishIsles.lua
--	AUTHOR:  HandyVac
------------------------------------------------------------------------------
-- Adapted from the AustraliaMap.lua used in the Australia scenario.

--[[
	Map Size reference

			  Civ V       Civ VI 		British Isles		Albion & Hibernia
	 
	Dual       40x24       44x26 		
	Tiny       55x36       60x36 		40x49				50x38
	Small      66x42       74x46 		50x61				60x45
	Standard   80x52       84x54 		60x73				70x53
	Large     104x64       96x60				
	Huge      128x80      106x66  		
	Enormous              128x80		
	Giant     180x94      180x94
	Ludicrous N/A        230x115 
--]]


include "MapEnums"
include "MapUtilities"
include "MountainsCliffs"
include "RiversLakes"
include "FeatureGenerator"
include "TerrainGenerator"
include "NaturalWonderGenerator"
include "ResourceGenerator"
include "AssignStartingPlots"

-- global values, mostly loaded from config
local g_iW, g_iH;
local g_iFlags = {};
local g_continentsFrac = nil;
local g_iNumTotalLandTiles = 0; 


local g_sMapType = "";	
local iMapTypeID = MapConfiguration.GetValue("HV_Maps_MapType");
	
if iMapTypeID == 0 then 
	g_sMapType = "BRITISH_ISLES";	
elseif iMapTypeID == 1 then 
	g_sMapType = "BRIT_POTOLOMY";	
end

print("g_sMapType is " .. g_sMapType);

-- can't get value for this until later, will be populated in GetMapInitData.
local g_sMapSize = "";
	
local g_bRegionalNatWonders = MapConfiguration.GetValue("HV_Maps_RegionalWonderPlacement");

-- -------------------------------------------------------------------
-- Input a Hash; Export width, height, and wrapX
function GetMapInitData(MapSize)
	-- height/width on HV maps are different from vanilla map sizes.
	-- look up values from custom sql table.
	
	
	local sMapType = "";	
	local iMapTypeID = MapConfiguration.GetValue("HV_Maps_MapType");
	
	if iMapTypeID == 0 then 
		sMapType = "BRITISH_ISLES";	
	elseif iMapTypeID == 1 then 
		sMapType = "BRIT_POTOLOMY";	
	end
	
	local sMapSize = "";
	local iWidth = 0;
	local iHeight = 0;
	
	
	-- turn the input hash into a string denoting name of map size.
	for row in GameInfo.Maps() do
		if(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_HUGE" then
			sMapSize = "MAPSIZE_HUGE"
		elseif(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_LARGE" then
			sMapSize = "MAPSIZE_LARGE"
		elseif(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_STANDARD" then
			sMapSize = "MAPSIZE_STANDARD"
		elseif(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_SMALL" then
			sMapSize = "MAPSIZE_SMALL"
		elseif(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_TINY" then
			sMapSize = "MAPSIZE_TINY"
		elseif(MapSize == row.Hash) and row.MapSizeType == "MAPSIZE_DUEL" then
			sMapSize = "MAPSIZE_DUEL"
		else
			--print("ERROR: unsupported map size.");
		end
	end
	
	-- set global value so other functions can refer to it
	g_sMapSize = sMapSize;
	print("g_sMapSize is " .. g_sMapSize);
	
	-- look through table to find entry for this map size & type.
	for Row in GameInfo.HV_Map_Sizes() do
		
		if Row.MapType == sMapType and Row.MapSize == sMapSize then
			iWidth = Row.Width;
			iHeight = Row.Height;
		end
	end
	-- if we didn't find a matching entry, then something went wrong.
	if iWidth == 0 and iHeight == 0 then
		print("ERROR: no valid HV_Map_Sizes entry for " .. sMapType .. " " .. sMapSize);
	end
	
	local WrapX = false;
	return {Width = iWidth, Height = iHeight, WrapX = WrapX,}
end
-------------------------------------------------------------------------------


function GenerateMap()
	print("Generating British Isles Map");
	local pPlot;

	-- Set globals
	g_iW, g_iH = Map.GetGridSize();
	g_iFlags = TerrainBuilder.GetFractalFlags();
	local temperature = 0;
	
	-- set plot types, e.g. hardcoded land/water for coastline.
	plotTypes = GenerateLandPlotsFromDatabase(g_sMapType, g_sMapSize);
		
	-- set terrain types, e.g. hills, plains, grass, etc.
	terrainTypes = GenerateTerrainTypesBritishIsles(plotTypes, g_iW, g_iH, g_iFlags, true, g_sMapType, g_sMapSize);
	
	-- apply the table of terrain types to the map
	for iPlay = 0, (g_iW * g_iH) - 1, 1 do
		pPlot = Map.GetPlotByIndex(iPlay);
		TerrainBuilder.SetTerrainType(pPlot, terrainTypes[iPlay]);
	end
	
	
	-- ====== Features ======

	AreaBuilder.Recalculate();
	local biggest_area = Areas.FindBiggestArea(false);
	print("After Adding Hills: ", biggest_area:GetPlotCount());

	-- River generation is affected by plot types, originating from highlands and preferring to traverse lowlands.
	AddRivers();
	
	-- Lakes would interfere with rivers, causing them to stop and not reach the ocean, if placed any sooner.
	local numLargeLakes = math.ceil(GameInfo.Maps[Map.GetMapSize()].Continents * 1.5);
	AddLakes(numLargeLakes);

	-- add features such as forest and marsh.
	AddFeatures();
	
	-- add cliffs.
	HV_AddCliffs(terrainTypes);
	
	-- add natural wonders
	local iNumWondersToPlace = GameInfo.Maps[Map.GetMapSize()].NumNaturalWonders;
	if g_bRegionalNatWonders then
		print(" ## Placing Natural Wonders by Region.")
		HV_PlaceNaturalWondersByRegion(iNumWondersToPlace);
	else
		print(" ## Placing Natural Wonders. Vanilla method.")
		local args = {
			numberToPlace = iNumWondersToPlace;
		};
		
		local nwGen = NaturalWonderGenerator.Create(args);
	end
	AreaBuilder.Recalculate();
	TerrainBuilder.AnalyzeChokepoints();
	TerrainBuilder.StampContinents();
	
	-- ====== Resources ========

	resourcesConfig = MapConfiguration.GetValue("resources");
	local args = {
		resources = resourcesConfig,
		LuxuriesPerRegion = 7,
	}
	local resGen = ResourceGenerator.Create(args);

	-- ====== Start Locations ========
	HV_DoRegionStartLocations();
	
	-- ===== tribal villages ======
	local GoodyGen = AddGoodies(g_iW, g_iH);
	
end


-- -------------------------------------------------------------------
-- sets start positions for all major and minor players, according to map regions.
function HV_DoRegionStartLocations()
	
	
	-- ======= Calculate Start Plots by Region========
	print("## Calculating start positions")
	
	-- get a list of region-names from the map SQL.
	local regionNames = {};
	local iNumRegions = 1;
	
	for Row in GameInfo.HV_Map_LandPlots() do
		if Row.MapType == g_sMapType and Row.MapSize == g_sMapSize then
			
			-- if this region is not already in the list, add it.
			if TestMembership(regionNames, Row.Region) == false then
				regionNames[iNumRegions] = Row.Region;
				iNumRegions = iNumRegions +1;
			end
			
		end
	end
	
	print("found " .. (iNumRegions-1) .. " Regions");
	
	-- get a list of all plots in each region.
	local iAreaSizeMin = 25;
	local iAreaSizeMax = 60;
		
	local minorStartRegions = {};
	local iNumMinorStarts = 1;
		
	local majorStartRegions = {};
	local iNumMajorStarts = 1;
	
	local bonusStartRegions = {};
	local iNumBonusStarts = 1;
	
	local RegionStartPlots = {};
	
	-- loop through each region.
	for iPlay = 1, iNumRegions-1, 1 do
		local sRegion = regionNames[iPlay];
		
		local plotList = {};
		local iNumPlots = 0;
		
		-- loop through all landstrips for this region in sql
		for Row in GameInfo.HV_Map_LandPlots() do
			-- if strip is for this region, add all plots in it.
			if Row.Region == sRegion and Row.MapType == g_sMapType and Row.MapSize == g_sMapSize then
				for iX = Row.Xstart, Row.Xend do
					local iPlay = Row.Y * g_iW + iX;
					local pPlot = Map.GetPlotByIndex(iPlay);
					
					-- don't count this plot if it's impassable (iPlay.e. a mountain)
					if pPlot:IsImpassable() == false then
						plotList[iNumPlots] = pPlot;
						iNumPlots = iNumPlots +1;
					end
				end
			end
		end
		
		
		--print(sRegion .. " contains " .. iNumPlots .. " passable plots");
		
		-- number of plots dictates whether this region is suitable for a major or minor civ.
				
		-- too small for a major civ, give it to a minor instead
		if iNumPlots > 0 and iNumPlots < iAreaSizeMin then 
		
			-- try to pick a random plot.
			-- brute-force multiple attempts to try and get one far away from other start plots.
			local pStartPlot = Map.GetPlot(0, 0);
			local iNumAttempts = 1;
			
			for iAttempt = 1, iAreaSizeMax, 1 do
				local iRand = TerrainBuilder.GetRandomNumber(iNumPlots, "start location roll")
				pStartPlot = plotList[iRand];
				
				-- if plot we picked is far enough away from all other start plots, it's fine. Use this.
				if HV_TooCloseToAnyPlot(pStartPlot, RegionStartPlots, 6) == false then
					-- success
					break
				end
				iNumAttempts = iNumAttempts +1
				-- we failed too many times, give up and just use this one, even though it's too close.
			end
			
			-- save this plot and region in the relevant tables.
			minorStartRegions[iNumMinorStarts] = sRegion;
			iNumMinorStarts = iNumMinorStarts + 1;
			
			RegionStartPlots[sRegion] = pStartPlot;
			
			--print(" - " .. sRegion .. " start location: " .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. " (Minor)" .. "(" .. iNumAttempts .. " attempts)" );
			
		-- a good size for one major civ
		elseif iNumPlots >=iAreaSizeMin and iNumPlots < iAreaSizeMax then 
	
			local pStartPlot = Map.GetPlot(0, 0);
			local iNumAttempts = 1;
			
			for iAttempt = 1, iAreaSizeMax, 1 do
				local iRand = TerrainBuilder.GetRandomNumber(iNumPlots, "start location roll")
				pStartPlot = plotList[iRand];
				
				-- if plot we picked is far enough away from all other start plots, it's fine. Use this.
				if HV_TooCloseToAnyPlot(pStartPlot, RegionStartPlots, 6) == false then
					-- success
					break
				end
				iNumAttempts = iNumAttempts + 1;
				-- we failed too many times, give up and just use this one, even though it's too close.
			end
			
			-- save this plot and region in the relevant tables.
			majorStartRegions[iNumMajorStarts] = sRegion;
			iNumMajorStarts = iNumMajorStarts +1;
			
			RegionStartPlots[sRegion] = pStartPlot;
			
			
			--print(" - " .. sRegion .. " start location: " .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. " (Major)"  .. "(" .. iNumAttempts .. " attempts)" );
			
		-- big enough for more than one civ.
		elseif iNumPlots >= iAreaSizeMax then
		
			-- determine how many start plots to put in this region
			local iNumStartsThisRegion = math.ceil(iNumPlots/iAreaSizeMax);
			
			-- pick that many plots.
			for iPlay=1, iNumStartsThisRegion, 1 do
				-- new name for this sub-region
				local sSubRegion = sRegion .. "_" .. iPlay;
				
				local pStartPlot = Map.GetPlot(0, 0);
				local iNumAttempts = 1;
				
				for iAttempt = 1, iAreaSizeMax, 1 do
					local iRand = TerrainBuilder.GetRandomNumber(iNumPlots, "start location roll")
					pStartPlot = plotList[iRand];
					
					-- if plot we picked is far enough away from all other start plots, it's fine. Use this.
					if HV_TooCloseToAnyPlot(pStartPlot, RegionStartPlots, 6) == false then
						-- success
						break
					end
					iNumAttempts = iNumAttempts + 1;
					-- we failed too many times, give up and just use this one, even though it's too close.
				end
			
				-- save this plot and region in the relevant tables.
				if iPlay == 1 then -- put the first one in the major starts list
					majorStartRegions[iNumMajorStarts] = sSubRegion;
					iNumMajorStarts = iNumMajorStarts +1;
					RegionStartPlots[sSubRegion] = pStartPlot;
					
					--print(" - " .. sSubRegion .. " start location: " .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. " (Major)"  .. "(" .. iNumAttempts .. " attempts)" );
				
				else -- put all subsequent ones in the bonus starts list
					bonusStartRegions[iNumBonusStarts] = sSubRegion;
					iNumBonusStarts = iNumBonusStarts +1;
					RegionStartPlots[sSubRegion] = pStartPlot;
					
					--print(" - " .. sSubRegion .. " start location: " .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. " (Bonus)"  .. "(" .. iNumAttempts .. " attempts)" );
		
				end
			end	
		end
	end
	
	print("found start locations: " .. (iNumMinorStarts-1) .. " minor, " .. (iNumMajorStarts-1) .. " major, "  .. (iNumBonusStarts-1) .. " bonus.");
		
	local iNumMajorPlayers = PlayerManager.GetAliveMajorsCount();
	local iNumMinorPlayers = PlayerManager.GetAliveMinorsCount();
	local iNumPlayers = iNumMajorPlayers + iNumMinorPlayers;
	print("found " .. iNumPlayers .. " players: " .. iNumMinorPlayers .. " minor, " .. iNumMajorPlayers  .. " major.");
	
	-- note, first player is index 0;

	
	-- shuffle the lists of major and minor regions
	majorStartRegions = GetShuffledCopyOfTable(majorStartRegions);
	bonusStartRegions = GetShuffledCopyOfTable(bonusStartRegions);
	minorStartRegions = GetShuffledCopyOfTable(minorStartRegions);
	
	-- note: GetShuffledCopyOfTable will ignore index 0 of a table, so make sure input starts with index 1.

	--print("printing shuffled tables");
	--PrintContentsOfTable(majorStartRegions);
	--PrintContentsOfTable(bonusStartRegions);
	--PrintContentsOfTable(minorStartRegions);

	-- append the shuffled tables together, with majors at the top, bonus in middle, and minor ones at the bottom.
	local allStartRegions = {};
	local iNumAllStartRegions = (iNumMinorStarts-1) + (iNumMajorStarts-1) + (iNumBonusStarts-1);
	print(iNumAllStartRegions .. " start regions in total");
	print("- - - - - - - - - - - - - - - - - -");
	for iPlay=0, iNumAllStartRegions-1, 1 do
		if iPlay < (iNumMajorStarts-1) then
			allStartRegions[iPlay] = majorStartRegions[iPlay+1];
		elseif iPlay >= (iNumMajorStarts-1) and iPlay < ( (iNumBonusStarts-1) + (iNumMajorStarts-1) ) then
			allStartRegions[iPlay] = bonusStartRegions[iPlay+1 - (iNumMajorStarts-1)];
		else
			allStartRegions[iPlay] = minorStartRegions[iPlay+1 - ( (iNumBonusStarts-1) + (iNumMajorStarts-1) )];
		end
	end
	--print("printing combined shuffled table");
	--PrintContentsOfTable(allStartRegions);
	
	-- loop through all players
	
	if iNumPlayers > iNumAllStartRegions then
		print("Too Many Players! " .. iNumPlayers - iNumAllStartRegions + 1 .. " of them will not be placed.")
	end
	
	
	local alreadyChosenRegions = {};
	local iNumAlreadyChosenRegions = 0;
	
	local alreadyPlacedPlayers = {};
	--local iNumAlreadyPlacedPlayers = {};
	
	-- ===== Assign True Start Regions =====
		
	local bTrueStarts = MapConfiguration.GetValue("HV_Maps_TrueStartRegions");
	if bTrueStarts == false then 
		print("not doing True Start Regions, was turned off in advanced setup");
	else
		print("doing True Start Regions");
		-- loop through players to do true-start regions.	
		for iPlay=0, iNumPlayers-1, 1 do
			-- lookup this player
			local pPlayerConfig = PlayerConfigurations[iPlay];
			local sPlayerType = pPlayerConfig:GetCivilizationTypeName();
			local sLeaderType = pPlayerConfig:GetLeaderTypeName();
			
			local sChosenRegion = "";
			local sDesiredRegion = "";
			local sCivRegion = "";
			local sLeaderRegion = "";
			
			-- loop through the sql table of true start regions
			for Row in GameInfo.HV_Map_TrueStartRegion() do 
			
							
				-- is the row valid for this map and civ?
				if (Row.MapType == g_sMapType or Row.MapType == nil) and (Row.MapSize == g_sMapSize or Row.MapSize == nil) and Row.Civilization == sPlayerType and Row.Region ~= nil then
					
					-- is is specific to this leader? Or generic, for any leader of this civ?
					if Row.Leader == sLeaderType then
						sLeaderRegion = Row.Region;
					elseif Row.Leader == nil then
						sCivRegion = Row.Region;
					end
				end
				
			end
			-- leader-specific starts take precedence over civ-specific.
			
			if sLeaderRegion ~= "" then
				sDesiredRegion = sLeaderRegion;
				print(sPlayerType .. " (" .. sLeaderType .. ")" .. " looking for a Leader-Specific true-start in " .. sDesiredRegion);
			elseif sLeaderRegion == "" and sCivRegion ~= "" then
				sDesiredRegion = sCivRegion;
				print(sPlayerType .. " (" .. sLeaderType .. ")" .. " looking for a true-start in " .. sDesiredRegion);
			end
			
			-- if we found a true-start in the table for this civ/leader, now look for a region to match it.
			if sDesiredRegion ~= "" then
				
				
				-- look for an available region that matches either matches or is a sub-region of the desired true-start.
				for iReg=0,  (iNumAllStartRegions-1), 1 do
					-- if this region has already been take, skip it.	
					if TestMembership(alreadyChosenRegions, allStartRegions[iReg]) == false then
					
						-- does this region match desired true-start?
						-- it may be divided into sub-regions with numbers appended to them, so look for matches *including* the true-start string
						if string.find(allStartRegions[iReg], sDesiredRegion) ~= nil then
				
							-- give the player this region
							sChosenRegion =  allStartRegions[iReg];
							-- add it to the list of already-chosen regions, to stop another player getting it.
							alreadyChosenRegions[iNumAlreadyChosenRegions] = sChosenRegion;
							iNumAlreadyChosenRegions = iNumAlreadyChosenRegions +1;
							
							-- add player to list of already placed, so we don't try and assign them a start again later
							alreadyPlacedPlayers[iPlay] = true;
							
							-- assign the player the start location for this region.
							local pStartPlot = RegionStartPlots[sChosenRegion];
				
							if pStartPlot == nil then
								print("could not place player " .. iPlay .. ": " .. sPlayerType);
							else
								Players[iPlay]:SetStartingPlot(pStartPlot);
								print("Player " .. iPlay .. ": " .. sPlayerType .. " starting in " .. sChosenRegion .. " (" .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. ")");
							end
							break;
						end
					else
						--print(allStartRegions[iReg] .. " is already taken")
					end
				end				
			end
		end	
	end
	-- ======= Assign Random Start Regions ======
	
	-- now loop through again to do random starts
	print("doing Random Start Regions");
	for iPlay=0, iNumPlayers-1, 1 do
		-- if we already placed this player earler (i.e. with true-start) skip them here.
		if alreadyPlacedPlayers[iPlay] ~= true then
	
			-- lookup this player
			local pPlayerConfig = PlayerConfigurations[iPlay];
			local sPlayerType = pPlayerConfig:GetCivilizationTypeName();
			local sChosenRegion = "";
			-- if we haven't already picked a true start region, pick the next un-chosen region from the shuffled list.
			if sChosenRegion == nil or sChosenRegion == "" then
				-- loop through the shuffled list
				for iReg=0, (iNumAllStartRegions-1), 1 do
					-- if this region's not already taken by another player.
					if TestMembership(alreadyChosenRegions, allStartRegions[iReg]) == false then
						-- give the player this region
						sChosenRegion =  allStartRegions[iReg];
						-- add it to the list of already-chosen regions, to stop another player getting it
						alreadyChosenRegions[iNumAlreadyChosenRegions] = sChosenRegion;
						iNumAlreadyChosenRegions = iNumAlreadyChosenRegions +1;
						break;
					end
				end
			end
			
			-- assign the player the start location for this region.
			local pStartPlot = RegionStartPlots[sChosenRegion];
			
			if pStartPlot == nil then
				print("could not place player " .. iPlay .. ": " .. sPlayerType);
			else
				Players[iPlay]:SetStartingPlot(pStartPlot);
				print("Player " .. iPlay .. ": " .. sPlayerType .. " starting in " .. sChosenRegion .. " (" .. pStartPlot:GetX() .. "," .. pStartPlot:GetY() .. ")");
			end
		end
	end

end
-------------------------------------------------------------------------------
	
-- loads sql table of land strips, and applies them to the map.
function GenerateLandPlotsFromDatabase(sMapType, sMapSize)

	
	print("Generating Coastline for a " .. sMapSize .. " " .. sMapType);
	local plotTypes = {};

	-- Start with it all as water
	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do
			local iPlay = y * g_iW + x;
			local pPlot = Map.GetPlotByIndex(iPlay);
			plotTypes[iPlay] = g_PLOT_TYPE_OCEAN;
			TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_OCEAN);
		end
	end

	-- plot out where the land should be.
	
	-- Remember: 00 is *bottom left* corner of map
	
	-- Each land strip is defined by: Y, X Start, X End

	local landStrips = {}
	
	-- loop through the database table of land strips.
	for Row in GameInfo.HV_Map_LandPlots() do
		-- is the entry for this mapsize?
		if Row.MapType == sMapType and Row.MapSize == sMapSize then
			-- add this strip of land to the map.
			for iX = Row.Xstart, Row.Xend do
				local iPlay = Row.Y * g_iW + iX;
				local pPlot = Map.GetPlotByIndex(iPlay);
				plotTypes[iPlay] = g_PLOT_TYPE_LAND;
				TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_GRASS); 
				g_iNumTotalLandTiles = g_iNumTotalLandTiles + 1;
			end
		end	
	end
	
	-- error checking
	if g_iNumTotalLandTiles <= 0 then
		print("Error. Couldn't find any land strips in HV_Map_LandPlots for " .. sMapType .. " " .. sMapSize);
	end

	-- calculate areas
	AreaBuilder.Recalculate();
	
	-- return list of all plots.
	return plotTypes;
end




function AddFeatures()
	print("Adding Features");

	-- get values from Advanced Setup screen.	
	local iConfigForestAmount =  MapConfiguration.GetValue("BritForestPercentage");
	if iConfigForestAmount == 9 then
		iConfigForestAmount = TerrainBuilder.GetRandomNumber(4, "Brit Forest Roll")
	end
	
	
	local iConfigMarshAmount =  MapConfiguration.GetValue("BritMarshPercentage");
	if iConfigMarshAmount == 9 then
		iConfigMarshAmount = TerrainBuilder.GetRandomNumber(4, "Brit Marsh Roll")
	end
	
	
	-- convert those values into the desired percentages
	if iConfigForestAmount == 4 then -- very abundant
		iConfigForestAmount = 60;
	elseif iConfigForestAmount == 3 then -- abundant
		iConfigForestAmount = 40;
	elseif iConfigForestAmount == 2 then -- standard
		iConfigForestAmount = 25;
	elseif iConfigForestAmount == 1 then --sparse
		iConfigForestAmount = 10;
	elseif iConfigForestAmount == 0 then -- none
		iConfigForestAmount = 0;
	end
	
	if iConfigMarshAmount == 4 then -- very abundant
		iConfigMarshAmount = 45;
	elseif iConfigMarshAmount == 3 then -- abundant
		iConfigMarshAmount = 25;
	elseif iConfigMarshAmount == 2 then -- standard
		iConfigMarshAmount = 10;
	elseif iConfigMarshAmount == 1 then -- sparse
		iConfigMarshAmount = 5;
	elseif iConfigMarshAmount == 0 then -- none
		iConfigMarshAmount = 0;
	end
	
	print("Config called for " .. iConfigForestAmount .. " percent forest cover");
	print("Config called for " .. iConfigMarshAmount .. " percent marsh cover");
	
	
	-- pass values to the feature generator.
	local args = {rainfall = 2, iForestPercent = iConfigForestAmount, iMarshPercent = iConfigMarshAmount}
	local featuregen = FeatureGenerator.Create(args);

	featuregen:AddFeatures();
	

end
------------------------------------------------------------------------------


function GenerateTerrainTypesBritishIsles(plotTypes, iW, iH, iFlags, bNoCoastalMountains, sMapType, sMapSize)
	print("Generating Terrain Types for a " .. sMapSize .. " " .. sMapType);
	
	local terrainTypes = {};
		
	-- ====== Fractals =======
	local fracXExp = -1;
	local fracYExp = -1;
	local grain_amount = 4;

	
	-- this fractal will be used to determine elevation of land
	highlandFrac = Fractal.Create(iW, iH, 
									grain_amount, iFlags, 
									fracXExp, fracYExp);
									
	-- thresholds for different catagories of elevation.
	-- TO DO: advanced setup options to tweak these. e.g. Hot, moderate or average climate/temperatre. large, medium or small mountain ranges. 
	-- maybe mountains ranges option could affect diameter of splats instead of fractal thresholds.
	
	iHighlandTop = highlandFrac:GetHeight(90); -- higher number will mean more mountains
	--print("iHighlandTop = " .. iHighlandTop);
	
	iHighlandMountain = highlandFrac:GetHeight(60);	-- lower number will mean more mountains
	--print("iHighlandMountain = " .. iHighlandMountain);
	
	iHighlandHill = highlandFrac:GetHeight(30);	-- higher number will mean more hills
	--print("iHighlandHill = " .. iHighlandHill);
	
	iHighlandFloor = highlandFrac:GetHeight(20); -- lower number will mean more hills
	--print("iHighlandFloor = " .. iHighlandFloor);
	
	-- this fractal will be used to determine climate
	tempFrac = Fractal.Create(iW, iH, 
									grain_amount, iFlags, 
									fracXExp, fracYExp);
	
	
	iTempGrass = tempFrac:GetHeight(20); -- lower number will mean more grass
	print("iTempGrass = " .. iTempGrass);
	
	iTempTundra = tempFrac:GetHeight(4);	-- higher number will mean more tundra
	print("iTempTundra = " .. iTempTundra);
	
	-- ======= Table Data ==============
	-- load info from sql tables to determine where to put mountain ranges
	
	print("## Looking for mountain splats");
	
	--  get a list of highland plots for this map from the sql database.
	local highandPlots = {}
	local iNumHighlandPlots = 0;
	
	-- look at all splats in the database table
	for Row in GameInfo.HV_Map_MountainSplats() do	
		--print("Row.MapType=" .. Row.MapType .. " Row.MapSize=" .. Row.MapSize .. " Row.X=" .. Row.X .. " Row.Y=" .. Row.Y);
	
		-- check splat is for correct map type and size
		if Row.MapType == sMapType and Row.MapSize == sMapSize then
			print("Adding mountain splat. Center: " .. Row.X .. "," .. Row.Y .. " Radius: " .. Row.Radius);
			
			-- test all plots in a square around the center plot
			for testX = (Row.X - Row.Radius), (Row.X + Row.Radius) do
				for testY = (Row.Y - Row.Radius), (Row.Y + Row.Radius) do
					-- is the plot within desired distance of the center?
					if Map.GetPlotDistance(testX, testY, Row.X, Row.Y) < Row.Radius then
						-- add plot to the table of results
						highandPlots[iNumHighlandPlots] = Map.GetPlot(testX, testY);
						iNumHighlandPlots = iNumHighlandPlots + 1;
					end
				end
			end
			
		end
	end
	
							
	
	-- ===== Elevation ======
	
	
	-- set elevation of land. E.g. hills and mountains.
	for iX = 0, iW - 1 do
		for iY = 0, iH - 1 do
			local index = (iY * iW) + iX;
			-- do nothing if this isn't a land tile.
			if (plotTypes[index] ~= g_PLOT_TYPE_OCEAN) then
			
				-- ===== Elevation ======								
				-- value of elevation fractal for this plot.
				local iHighlandVal = highlandFrac:GetHeight(iX, iY);
				
				local iPlotElevation = 0;
				local iPlotClimate = 0;
				local bIsCliff = false;
				
				local pPlot = Map.GetPlot(iX, iY);
				
				-- if this plot is in an area we want to be moutainous, use the full fractal value
				if TestMembership(highandPlots, pPlot) == true then
					-- do nothing				
				else
					-- if plot is outside of a mountainous area, flatten it by reducing the fractal value.
					iHighlandVal = iHighlandVal/2;
				end
					
				
				-- very highest points on fractal are hills, not mountains. Hopefully will prevent too big clumps of continuous mountain.
				if (iHighlandVal >= iHighlandTop) then
					iPlotElevation = 1;
				
				-- high points on fractal are mountains.
				elseif (iHighlandVal >=  iHighlandMountain) and (iHighlandVal < iHighlandTop) then 
					iPlotElevation = 2;
					
				-- points lower than that are hills
				elseif (iHighlandVal >=  iHighlandHill) and (iHighlandVal < iHighlandMountain) then
					iPlotElevation =  1;
				
				-- and the lowest points are flat ground				
				else
					iPlotElevation = 0;
					
				end
				
				
					
				-- ===== Temperature north/south ======
				
				-- get temperature fractal value for this tile.
				-- use only a percentage of it, depending on how far north we are.
				local iNorthPercent = (iY/iH) * 100;
				local iTempVal = (tempFrac:GetHeight(iX, iY)/100) * (100-iNorthPercent);
				
				if iTempVal >= iTempGrass then
					iPlotClimate = 2; -- grass
				elseif iTempVal < iTempGrass and iTempVal >= iTempTundra then
					iPlotClimate = 1; -- plains
				elseif iTempVal < iTempTundra then
					iPlotClimate = 0; -- tundra
				end
				
				-- ===== Set Terrain and plot Types  ======
				
				-- Assemble climate and elevation values into an actual terrain type.
				if iPlotElevation >= 2 then
					-- mountain, but what climate?
					if iPlotClimate >= 2 then -- warm
						terrainTypes[index] = g_TERRAIN_TYPE_GRASS_MOUNTAIN;
					elseif iPlotClimate == 1 then -- moderate
						terrainTypes[index] = g_TERRAIN_TYPE_PLAINS_MOUNTAIN;
					elseif iPlotClimate <=0 then -- cold
						terrainTypes[index] = g_TERRAIN_TYPE_TUNDRA_MOUNTAIN;
					end
					
				elseif iPlotElevation == 1 then
					-- hill, but what climate?
					if iPlotClimate >= 2 then -- warm
						terrainTypes[index] = g_TERRAIN_TYPE_GRASS_HILLS;
					elseif iPlotClimate == 1 then -- moderate
						terrainTypes[index] = g_TERRAIN_TYPE_PLAINS_HILLS;
					elseif iPlotClimate <=0 then -- cold
						terrainTypes[index] = g_TERRAIN_TYPE_TUNDRA_HILLS;
					end
				
				elseif iPlotElevation <= 0 then
					-- flat, but what climate?
					if iPlotClimate >= 2 then -- warm
						terrainTypes[index] = g_TERRAIN_TYPE_GRASS;
					elseif iPlotClimate == 1 then -- moderate
						terrainTypes[index] = g_TERRAIN_TYPE_PLAINS;
					elseif iPlotClimate <=0 then -- cold
						terrainTypes[index] = g_TERRAIN_TYPE_TUNDRA;
					end
					
				end
				
			end
		end
	end
	
	
	-- TO DO: load temperature variation and allowable climates from SQL instead.
	
	-- ===== Coast and Ocean =====
	
	
	
	-- turn all ocean adjacent to land into coastal water
	for iX = 0, iW - 1 do
		for iY = 0, iH - 1 do
			local index = (iY * iW) + iX;
			if (plotTypes[index] == g_PLOT_TYPE_OCEAN) then
				if (IsAdjacentToLand(plotTypes, iX, iY)) then
					terrainTypes[index] = g_TERRAIN_TYPE_COAST;
				else
					terrainTypes[index] = g_TERRAIN_TYPE_OCEAN;
				end
			end
		end
	end

	-- turn some of the ocean adjacent to coastal water into more coastal water
	print("Expanding coasts");
	for iI = 0, 2 do
		local shallowWaterPlots = {};
		for iX = 0, iW - 1 do
			for iY = 0, iH - 1 do
				local index = (iY * iW) + iX;
				if (terrainTypes[index] == g_TERRAIN_TYPE_OCEAN) then
					-- Chance for each eligible plot to become an expansion is 1 / iExpansionDiceroll.
					-- Default is two passes at 1/4 chance per eligible plot on each pass.
					if (IsAdjacentToShallowWater(terrainTypes, iX, iY) and TerrainBuilder.GetRandomNumber(4, "add shallows") == 0) then
						table.insert(shallowWaterPlots, index);
					end
				end
			end
		end
		for index, plotID in ipairs(shallowWaterPlots) do
			terrainTypes[plotID] = g_TERRAIN_TYPE_COAST;
		end
	end
	
	return terrainTypes; 
end


------------------------------------------------------------------------------
-- no ice or jungles wanted, so replace the default functions for adding them.
function FeatureGenerator:AddIceAtPlot(plot, iX, iY, lat)
	return;
end

function FeatureGenerator:AddJunglesAtPlot(plot, iX, iY, lat)
	return false;
end

-- change the default forest-adding function. 
-- this is a smaller scale map than most, so we want bigger clumps of forest than the vanilla generator creates
function FeatureGenerator:AddForestsAtPlot(plot, iX, iY)
		
	-- check if this plot's terrain is even suitable for a forest.
	if(TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FOREST)) then
		-- now check if the map already has too many forests.
		if(math.ceil(self.iForestCount * 100 / self.iNumLandPlots) <= self.iForestMaxPercent) then
	
			-- if all is well, check how many forest are adjacent to this plot.
			local iAdjacent = TerrainBuilder.GetAdjacentFeatureCount(plot, g_FEATURE_FOREST);
			
			-- if there are exactly three other forests adjacent to this, always add a forest
			if iAdjacent == 3 then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST);
				self.iForestCount = self.iForestCount + 1;
				return true;
				
			-- if there are more than three forests adjacent, never add one
			elseif iAdjacent > 3 then
			
			-- if there are fewer than three forests adjacent, 50% chance to add one.
			elseif iAdjacent < 3 then
				if TerrainBuilder.GetRandomNumber(100, "Roll for British Forest") > 50 then
					TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST);
					self.iForestCount = self.iForestCount + 1;
					return true;
				end
			end
		end
	end
	
	return false;
end

-- same for marshes, change rules to make them appear in larger clumps.
function FeatureGenerator:AddMarshAtPlot(plot, iX, iY)
	--Marsh Check. First see if it can place the feature.
	
	-- check if this plot's terrain is even suitable for a forest.
	if(TerrainBuilder.CanHaveFeature(plot, g_FEATURE_MARSH)) then
		-- now check if the map already has too many marshes.
		if(math.ceil(self.iMarshCount * 100 / self.iNumLandPlots) <= self.iMarshMaxPercent) then
	
			-- if all is well, check how many marshes are adjacent to this plot.
			local iAdjacent = TerrainBuilder.GetAdjacentFeatureCount(plot, g_FEATURE_MARSH);
			
			-- if there are exactly three other marshes adjacent to this, always add one.
			if iAdjacent == 3 then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_MARSH);
				self.iMarshCount = self.iMarshCount + 1;
				return true;
				
			-- if there are more than three marshes adjacent, never add one
			elseif iAdjacent > 3 then
			
			-- if there are fewer than three marshes adjacent, 50% chance to add one.
			elseif iAdjacent < 3 then
				if TerrainBuilder.GetRandomNumber(100, "Roll for British marsh") > 50 then
					TerrainBuilder.SetFeatureType(plot, g_FEATURE_MARSH);
					self.iMarshCount = self.iMarshCount + 1;
					return true;
				end
			end
		end
	end

	-- return false if we didn't add anything.
	return false;
end


function HV_PlaceNaturalWondersByRegion(iNumToPlace)
	-- ==== assemble list of regions on map. ======
	local aRegionNames = {};
	local iNumRegions = 0;
	local iNumToPlace = GameInfo.Maps[Map.GetMapSize()].NumNaturalWonders;
	
	
	-- loop through list of landstrips, record every region name once.
	for Row in GameInfo.HV_Map_LandPlots() do
		if Row.MapType == g_sMapType and Row.MapSize == g_sMapSize then
			-- if this region is not already in the list, add it.
			if TestMembership(aRegionNames, Row.Region) == false then
				--aRegionNames[iNumRegions] = Row.Region;
				iNumRegions = iNumRegions +1;
				table.insert(aRegionNames, Row.Region);
			end
		end
	end
	
	print("found " .. iNumRegions .. " Regions");
	--PrintContentsOfTable(aRegionNames);
	
	-- ====== assemble list of valid wonders for this map. =======
	local aWonderNames = {}; -- will be a list of names of wonders 
	local aWonderNamesTruestart = {}; -- same as above, but for wonders which have a true-start region
	local iNumWonders = 0;
	
	-- loop through database table of features, find all wonders.
	for Row in GameInfo.Features() do
		if Row.NaturalWonder then
			--aWonderNames[iNumWonders] = Row.FeatureType;
			iNumWonders = iNumWonders +1;
			table.insert(aWonderNames, Row.FeatureType);
		end
	end
	print("found " .. iNumWonders .. " Natural Wonders in database.");
	--PrintContentsOfTable(aWonderNames);
		
	
	-- loop through regions, looking for valid natural wonder plots in each one.
	local aaValidWonderPlots = {}; -- will be a table of valid plots for wonders on this map, including plot, name of wonder and name of region.
	local aaValidWonderPlotsTrueStart = {};
	for index, sRegionName in ipairs(aRegionNames) do
		--print("looking for wonders valid in: " .. sRegionName);
		
		-- loop through all wonders
		for index, sWonderType in ipairs(aWonderNames) do
			local iNumPlotsFound = 0;
			
			local aValidRegions = HV_FeatureValidRegions(sWonderType);
			
			-- loop through all land plots in this region, to check for eligablility.
			for Row in GameInfo.HV_Map_LandPlots() do
				if Row.Region == sRegionName and Row.MapType == g_sMapType and Row.MapSize == g_sMapSize then
					for iX = Row.Xstart, Row.Xend do
						local iPlotID = Row.Y * g_iW + iX;
						local pPlot = Map.GetPlotByIndex(iPlotID);
						
						-- if wonder has custom placement rules, use those
						if GameInfo.Features[sWonderType].CustomPlacement then
							
							if CustomCanHaveFeature(pPlot, GameInfo.Features[sWonderType].Index) then
								local ValidPlot = {};
								ValidPlot.Plot = pPlot;
								ValidPlot.Wonder = sWonderType;
								ValidPlot.Region = sRegionName;
								
								-- if this is a true-start region, add to true-starts list
								if TestMembership(aValidRegions, sRegionName) then
									table.insert(aaValidWonderPlotsTrueStart, ValidPlot);
									iNumPlotsFound = iNumPlotsFound+1;
								else
									-- otherwise add to generic list
									table.insert(aaValidWonderPlots, ValidPlot);
									iNumPlotsFound = iNumPlotsFound+1;
								end
							end
						
						else -- otherwise use generic placement rules
							if TerrainBuilder.CanHaveFeature(pPlot, GameInfo.Features[sWonderType].Index, false) then
								--print(" - " .. sWonderType .. " is valid");
								local ValidPlot = {};
								ValidPlot.Plot = pPlot;
								ValidPlot.Wonder = sWonderType;
								ValidPlot.Region = sRegionName;
								
								-- if this is a true-start region, add to true-starts list
								if TestMembership(aValidRegions, sRegionName) then
									table.insert(aaValidWonderPlotsTrueStart, ValidPlot);
									iNumPlotsFound = iNumPlotsFound+1;
								else
									-- otherwise add to generic list
									table.insert(aaValidWonderPlots, ValidPlot);
									iNumPlotsFound = iNumPlotsFound+1;
								end
							end
						end
					end
				end
			end
			
			-- only land plots have regions, so need to do water plots seperately.
			
			-- loop through all ocean plots on the map, to check for eligability
			for iX = 0, g_iW - 1 do
				for iY = 0, g_iH - 1 do
					local iPlotID = iY * g_iW + iX;
					local pPlot = Map.GetPlotByIndex(iPlotID);
					
					if pPlot:IsWater() then
						-- if wonder has custom placement rules, use those
						if GameInfo.Features[sWonderType].CustomPlacement then
							
							if CustomCanHaveFeature(pPlot, GameInfo.Features[sWonderType].Index) then
								local ValidPlot = {};
								ValidPlot.Plot = pPlot;
								ValidPlot.Wonder = sWonderType;
								ValidPlot.Region = "OCEAN";
								
								-- there are no true-start regions for ocean, add to generic list
								table.insert(aaValidWonderPlots, ValidPlot);
								iNumPlotsFound = iNumPlotsFound+1;
								
							end
						
						else -- otherwise use generic placement rules
							if TerrainBuilder.CanHaveFeature(pPlot, GameInfo.Features[sWonderType].Index, false) then
								--print(" - " .. sWonderType .. " is valid");
								local ValidPlot = {};
								ValidPlot.Plot = pPlot;
								ValidPlot.Wonder = sWonderType;
								ValidPlot.Region = "OCEAN";
								
								-- there are no true-start regions for ocean, add to generic list
								table.insert(aaValidWonderPlots, ValidPlot);
								iNumPlotsFound = iNumPlotsFound+1;
								
							end
						end
					end
				end
			end
			
			--[[
			if iNumPlotsFound > 0 then
				print(" - found " .. iNumPlotsFound .. " valid plots for " .. sWonderType);
			end
			--]]
		end
		
	end

	-- ====== shuffle tables. ========
	print(#aaValidWonderPlots .. " valid wonder plots in total");
	
	aRegionNames = GetShuffledCopyOfTable(aRegionNames);
	aWonderNames = GetShuffledCopyOfTable(aWonderNames);
	aaValidWonderPlots = GetShuffledCopyOfTable(aaValidWonderPlots);
	aaValidWonderPlotsTrueStart = GetShuffledCopyOfTable(aaValidWonderPlotsTrueStart);
	
	-- append the shuffled tables together, so true-starts are at the top.
	local aaAllValidWonderPlots = {}
	for index, Row in ipairs(aaValidWonderPlotsTrueStart) do 
		table.insert(aaAllValidWonderPlots, Row);
	end
	for index, Row in ipairs(aaValidWonderPlots) do
		table.insert(aaAllValidWonderPlots, Row);
	end
	

	
	local aPlacedWonders = {}
	local aPlacedPlots = {}
	local aPlacedRegions = {}
	local iMinDist = 10;
	local aPlacedWonderIDs = {}
	
	-- go through the shuffled list and place the first wonders we find.
	for index, Row in ipairs(aaAllValidWonderPlots) do
		-- skip this if enough wonders have been placed already
		if #aPlacedWonders <  iNumToPlace then
			
			-- has this wonder or region been placed already?
			if TestMembership(aPlacedWonders, Row.Wonder) == false and TestMembership(aPlacedRegions, Row.Region) == false then
				-- is it far enough away from plots that already contain wonders?
				if HV_TooCloseToAnyPlot(Row.Plot, aPlacedPlots, iMinDist) == false then
					-- we can place this wonder here.
					
					-- add it to the lists, so we don't try and place same thing again.
					table.insert(aPlacedWonders, Row.Wonder);
					table.insert(aPlacedPlots, Row.Plot);
					table.insert(aPlacedRegions, Row.Region);
					
					-- place it on the map.
					local eWonderID = GameInfo.Features[Row.Wonder].Index;
					local pPlot = Row.Plot;
					table.insert(aPlacedWonderIDs, eWonderID);
						
					-- use custom placement rules if neccesary
					if GameInfo.Features[Row.Wonder].CustomPlacement then
						CustomSetFeatureType(pPlot, eWonderID);
								
					else -- otherwise use generic placement rules
					
						TerrainBuilder.SetFeatureType(pPlot, eWonderID);

						ResetTerrain(pPlot:GetIndex());

						local plotX = pPlot:GetX();
						local plotY = pPlot:GetY();

						for dx = -2, 2 do
							for dy = -2,2 do
								local otherPlot = Map.GetPlotXY(plotX, plotY, dx, dy, 2);
								if(otherPlot) then
									if(otherPlot:IsNaturalWonder() == true) then
										ResetTerrain(otherPlot:GetIndex());
									end
								end
							end
						end
						
						
					end
					
					print("placed " .. Row.Wonder .. " in " .. Row.Region .. " (" .. Row.Plot:GetX() .. "," .. Row.Plot:GetY() .. ")" );
				
				end
				
			end	
		end
	end
	
	print("placed " .. #aPlacedWonders .. " natural wonders in total");
	
	-- TO DO: place remaining wonders by vanilla method
	--[[
	print(" ## Placing " .. iNumToPlace - #aPlacedWonders .. " more Natural Wonders by Vanilla method.")
	print(" Invalid List: " );
	PrintContentsOfTable(aPlacedWonderIDs);
	PrintContentsOfTable(aPlacedWonders);
	local args = {
		numberToPlace = iNumToPlace - #aPlacedWonders,
		Invalid = aPlacedWonderIDs,
	}; -- invalid list does not seem to work, for some reason? Ends up placing duplicates of already-placed wonders.

	local nwGen = NaturalWonderGenerator.Create(args);
	
	--]]
end


-- had to fix a bug in vanilla version of this, so it doesn't crash when trying to place g_FEATURE_PIOPIOTAHI adjacent to the edge of the map. 
function CustomGetMultiTileFeaturePlotList(pPlot, eFeatureType, aPlots)

	-- First check this plot itself
	if (not TerrainBuilder.CanHaveFeature(pPlot, eFeatureType, true)) then
		return false;
	else
		table.insert(aPlots, pPlot:GetIndex());
	end

	-- Which type of custom placement is it?
	local customPlacement = GameInfo.Features[eFeatureType].CustomPlacement;

		-- 2 tiles inland, east-west facing camera
	if (customPlacement == "PLACEMENT_TORRES_DEL_PAINE" or
	    customPlacement == "PLACEMENT_YOSEMITE") then

		-- Assume first tile is the western one, check the one to the east
		local pAdjacentPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (pAdjacentPlot ~= nil and TerrainBuilder.CanHaveFeature(pAdjacentPlot, eFeatureType, true) == true) then
			table.insert(aPlots, pAdjacentPlot:GetIndex());
			return true;
		end

	-- 2 tiles on coast, roughly facing camera
	elseif (customPlacement == "PLACEMENT_CLIFFS_DOVER") then
		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);

		-- W and SW are water, see if SE works
		local pSecondPlot;
		if (pWPlot ~= nil and pSWPlot ~= nil and pWPlot:IsWater() and pWPlot:IsLake() == false and pSWPlot:IsWater() and pWPlot:IsLake() == false) then
			pSecondPlot = pSEPlot;

		-- SW and SE are water, see if E works
		elseif (pSWPlot ~= nil and pSEPlot ~= nil and pSWPlot:IsWater() and pSWPlot:IsLake() == false and pSEPlot:IsWater() and pSEPlot:IsLake() == false) then
			pSecondPlot = pEPlot;

		-- SE and E are water, see if NE works
		elseif (pSWPlot ~= nil and pEPlot ~= nil  and pSEPlot:IsWater() and pSEPlot:IsLake() == false and pEPlot:IsWater() and pEPlot:IsLake() == false) then
			pSecondPlot = pNEPlot;
		
		else
			return false;
		end

		if (TerrainBuilder.CanHaveFeature(pSecondPlot, eFeatureType, true)) then
			table.insert(aPlots, pSecondPlot:GetIndex());
			return true;
		end

	-- 2 tiles, one on coastal land and one in water, try to face camera if possible
	elseif (customPlacement == "PLACEMENT_GIANTS_CAUSEWAY") then

		-- Assume first tile a land tile without hills, check around it in a preferred order for water
		if (pPlot:IsWater() or pPlot:IsHills()) then
			return false;
		end

		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		if (pSWPlot ~= nil and pSWPlot:IsWater()) then
			table.insert(aPlots, pSWPlot:GetIndex());
			return true;
		end

		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		if (pSEPlot ~= nil and pSEPlot:IsWater()) then
			table.insert(aPlots, pSEPlot:GetIndex());
			return true;
		end

		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		if (pWPlot ~= nil and pWPlot:IsWater()) then
			table.insert(aPlots, pWPlot:GetIndex());
			return true;
		end

		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (pEPlot ~= nil and pEPlot:IsWater()) then
			table.insert(aPlots, pEPlot:GetIndex());
			return true;
		end

		local pNWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
		if (pNWPlot ~= nil and pNWPlot:IsWater()) then
			table.insert(aPlots, pNWPlot:GetIndex());
			return true;
		end

		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		if (pNEPlot ~= nil and pNEPlot:IsWater()) then
			table.insert(aPlots, pNEPlot:GetIndex());
			return true;
		end
		
	-- 4 tiles (triangle plus a tail)
	elseif (customPlacement == "PLACEMENT_RORAIMA") then

		-- This one does require three in a row, so let's find that first
		for i = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
			local pFirstPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), i);
			if (pFirstPlot ~= nil and TerrainBuilder.CanHaveFeature(pFirstPlot, eFeatureType, true)) then
				local pSecondPlot = Map.GetAdjacentPlot(pFirstPlot:GetX(), pFirstPlot:GetY(), i);
				if (pSecondPlot ~= nil and TerrainBuilder.CanHaveFeature(pSecondPlot, eFeatureType, true)) then
					local iNewDir = i - 1;
					if iNewDir == -1 then
						iNewDir = 5;
					end
					local pThirdPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), iNewDir);
					if (pThirdPlot ~= nil and TerrainBuilder.CanHaveFeature(pThirdPlot, eFeatureType, true)) then
						table.insert(aPlots, pFirstPlot:GetIndex());
						table.insert(aPlots, pSecondPlot:GetIndex());
						table.insert(aPlots, pThirdPlot:GetIndex());
						return true;
					end
				end
			end
		end

	-- 3 tiles in a straight line
	elseif (customPlacement == "PLACEMENT_ZHANGYE_DANXIA") then

		for i = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
			local pFirstPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), i);
			if (pFirstPlot ~= nil and TerrainBuilder.CanHaveFeature(pFirstPlot, eFeatureType, true)) then
				local pSecondPlot = Map.GetAdjacentPlot(pFirstPlot:GetX(), pFirstPlot:GetY(), i);
				if (pSecondPlot ~= nil and TerrainBuilder.CanHaveFeature(pSecondPlot, eFeatureType, true)) then
					table.insert(aPlots, pFirstPlot:GetIndex());
					table.insert(aPlots, pSecondPlot:GetIndex());
					return true;
				end
			end
		end

	-- 3 tiles in triangle coast on front edge, land behind (with any rotation)
	elseif (customPlacement == "PLACEMENT_PIOPIOTAHI") then

		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		local pNWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		
		-- all 6 hexes around must be land

		if (pNWPlot == nil or pNEPlot == nil or pWPlot== nil or pSWPlot== nil or pSEPlot== nil or pEPlot== nil or pNWPlot:IsWater() or pNEPlot:IsWater() or pWPlot:IsWater() or pSWPlot:IsWater() or pSEPlot:IsWater() or pEPlot:IsWater()) then
			return false;
		else
			-- find two adjacent plots that can both serve for this NW
			local bWValid  = TerrainBuilder.CanHaveFeature(pWPlot, eFeatureType, true);
			local bNWValid = TerrainBuilder.CanHaveFeature(pNWPlot, eFeatureType, true);
			local bNEValid = TerrainBuilder.CanHaveFeature(pNEPlot, eFeatureType, true);
			local bEValid  = TerrainBuilder.CanHaveFeature(pEPlot, eFeatureType, true);
			local bSEValid = TerrainBuilder.CanHaveFeature(pSEPlot, eFeatureType, true);
			local bSWValid = TerrainBuilder.CanHaveFeature(pSWPlot, eFeatureType, true);

			if (bSEValid ~= nil and bSWValid ~= nil and bSEValid == true and bSWValid == true ) then
				pWaterCheck1 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				
				-- HV: Check if these adjacent plots exist, because they may not on my British Isles map.
				if (pWaterCheck1 == nil or pWaterCheck2 == nil or pWaterCheck3 == nil) then
					--print("ERROR: Tried to put g_FEATURE_PIOPIOTAHI next to edge of map!");
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pSEPlot:GetIndex());
					table.insert(aPlots, pSWPlot:GetIndex());
					return true;
				end
			end
			
			if (bEValid  ~= nil and bSEValid ~= nil and bEValid  == true and bSEValid == true ) then
				if (Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_EAST) == nil) then
					return false;
				elseif ( Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST) == nil) then
					return false;
				end

				pWaterCheck1 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);

				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pEPlot:GetIndex());
					table.insert(aPlots, pSEPlot:GetIndex());
					return true;
				end
			end
			
			if (bSWValid  ~= nil and bWValid ~= nil and bSWValid  == true and bWValid == true ) then
				if (Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST) == nil) then
					return false;
				elseif ( Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST) == nil) then
					return false;
				end

				pWaterCheck1 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pSWPlot:GetIndex());
					table.insert(aPlots, pWPlot:GetIndex());
					return true;
				end
			end

			if (bWValid  ~= nil and bNWValid ~= nil and bWValid  == true and bNWValid == true ) then
				if (Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_WEST) == nil) then
					return false;
				elseif ( Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST) == nil) then
					return false;
				end
				
				pWaterCheck1 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pWPlot:GetIndex());
					table.insert(aPlots, pNWPlot:GetIndex());
					return true;
				end
			end
			
			if (bNEValid  ~= nil and bEValid ~= nil and bNEValid  == true and bEValid == true ) then
				if (Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST) == nil) then
					return false;
				elseif ( Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST) == nil) then
					return false;
				end
				
				pWaterCheck1 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pNEPlot:GetIndex());
					table.insert(aPlots, pEPlot:GetIndex());
					return true;
				end
			end

			if (bNWValid  ~= nil and bNEValid ~= nil and bNWValid  == true and bNEValid == true ) then
				if (Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST) == nil) then
					return false;
				elseif (Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST) == nil) then
					return false;
				end

				pWaterCheck1 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pNWPlot:GetIndex());
					table.insert(aPlots, pNEPlot:GetIndex());
					return true;
				end
			end
		end
	end

	return false;
end

-- ---------------------------------------------------------------
-- --              Utility functions 
-- ---------------------------------------------------------------
-- accepts a table, returns the number of entries in that table
function GetTableLength (Table)
	local iCount = 0;

	if Table ~= nil then
		for Entry in pairs(Table) do 
			iCount = iCount + 1;
		end
	end

	return iCount;
end

-- accepts a plot, returns the land-strip region that plot is in.
function HV_GetRegionOfPlot(pInputPlot)
	-- get size of map.
	local sMapType = "";	
	local iMapTypeID = MapConfiguration.GetValue("HV_Maps_MapType");
	 
	
	local iInputX = pInputPlot:GetX();
	local iInputY = pInputPlot:GetY();
	
	if pInputPlot:IsWater() then 
		return "NO_REGION";
		
	else
		local sRegion = "FAILED";
		-- loop through all land-strips in HV_Map table
		for Row in GameInfo.HV_Map_LandPlots() do
			-- is row for this mapsize and type?
			if Row.MapType == g_sMapType and Row.MapSize == g_sMapSize then
				-- is plot on this Y?
				--print("Strip Y: " .. Row.Y .. " region: " .. Row.Region .. " X: " .. Row.Xstart .. " to " .. Row.Xend);
				if Row.Y == iInputY and Row.Region ~= nil then
					-- is it between the start and end X?
					if 	iInputX >= Row.Xstart and iInputX <= Row.Xend then
						-- return region of this land-strip
						sRegion =  Row.Region;
						--print("region " .. Row.Region);
					end
				end
			end	
		end
	
		--print("plot strip Y: " .. iInputY .. " X: " .. iInputX .. " is region " .. sRegion);
		return sRegion;
	
	end
	
	--TO DO: return water plots near to land as part of that region.
end

-- accepts x,y values, returns true if plot is land adjacent to water
function HV_IsAdjacentToWaterPlot(x, y)
	-- returns true is plot is adjacent to coast or ocean
	local plot = Map.GetPlot(x, y);
	if plot ~= nil then
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
			local testPlot = Map.GetAdjacentPlot(x, y, direction);
			if testPlot ~= nil then
				if testPlot:IsWater() == true then -- Adjacent plot is water
					return true
				end
			end
		end
	end
	return false
end

function HV_AddCliffs(terrainTypes)
	print("Adding Cliffs ");
	
	local iW, iH = Map.GetGridSize();
	local iMountainsToCliffs = 0;
	local iHillsToCliffs = 0;
	local iFlatToCliffs = 0;
	local iMountainsRemoved = 0;
	local iRoll = 0;
	

	
	-- load amount of cliffs from config.
	local eCliffsConfig = MapConfiguration.GetValue("BritCliffPercentage");
	
	if eCliffsConfig == 9 then
		eCliffsConfig = TerrainBuilder.GetRandomNumber(4, "Brit Cliff Roll")
	end
	
	local iCliffConfigPercentage = 0;
	-- convert those values into the desired percentages
	if eCliffsConfig == 4 then -- very abundant
		iCliffConfigPercentage = 60;
	elseif eCliffsConfig == 3 then -- abundant
		iCliffConfigPercentage = 40;
	elseif eCliffsConfig == 2 then -- standard
		iCliffConfigPercentage = 25;
	elseif eCliffsConfig == 1 then --sparse
		iCliffConfigPercentage = 10;
	elseif eCliffsConfig == 0 then -- none
		iCliffConfigPercentage = 0;
	end
	
	print("config called for " .. iCliffConfigPercentage .. " percent cliff coverage");
	
	local iFlatCliffPercent = iCliffConfigPercentage;
	local iHillsCliffPercent = iCliffConfigPercentage * 1.5;
	local iMountainsCliffsPercent = iCliffConfigPercentage * 2 ;
	
	
	for iX = 0, iW - 1 do
		for iY = 0, iH - 1 do
			local index = (iY * iW) + iX;
			local pPlot = Map.GetPlotByIndex(index);
			local area = pPlot:GetArea();
			
			-- check plot is coastal, and not a tiny island
			if HV_IsAdjacentToWaterPlot(iX, iY) then
				
				-- remove all coastal mountains, convert most of them to cliffs.
				if pPlot:IsMountain() then
					-- remove mountain
					TerrainBuilder.SetTerrainType(pPlot, (terrainTypes[index] - g_TERRAIN_BASE_TO_HILLS_DELTA) );
					iMountainsRemoved = iMountainsRemoved +1;
					
					-- set plot to cliff
					iRoll = TerrainBuilder.GetRandomNumber(100, "rolling to add cliffs");
				
					if iRoll < iMountainsCliffsPercent then
						SetCliff(terrainTypes, iX, iY);
						iMountainsToCliffs = iMountainsToCliffs +1;
					end
					
				end
				
				-- ensure tiny islands don't get entirely surrounded by cliffs.
				if area:GetPlotCount() > 1 and not area:HasNoFlatCoast() then 
				
					-- add some cliffs to coastal hills
					if pPlot:IsHills() then
						iRoll = TerrainBuilder.GetRandomNumber(100, "rolling to add cliffs");
						
						if iRoll < iHillsCliffPercent then 
							SetCliff(terrainTypes, iX, iY);
							iHillsToCliffs = iHillsToCliffs +1;
						end
					end
					
					-- add some cliffs to flat coastal land
					if pPlot:IsFlatlands() then
						iRoll = TerrainBuilder.GetRandomNumber(100, "rolling to add cliffs");
						
						if iRoll < iFlatCliffPercent then 
							SetCliff(terrainTypes, iX, iY);
							iFlatToCliffs = iFlatToCliffs +1;
						end
					end
				end
			end
		end
	end
	
	print("removed " .. iMountainsRemoved .. " coastal mountains, converted " .. iMountainsToCliffs .. " of them into cliffs");
	print("converted " .. iHillsToCliffs .. " coastal hills into cliffs");
	print("converted " .. iFlatToCliffs .. " coastal flats into cliffs");
	
	
end

-- accepts a plot, list of plots, and a distance. Returns true if first plot within that distance of any of the other plots.
function HV_TooCloseToAnyPlot (pTestPlot, plotList, iMinDistance)
	
	-- loop through all plots in the list
	for index, pPlot in pairs(plotList) do
		-- check distance from this plot to test plot.
		local iDistanceFromTestPlot = Map.GetPlotDistance ( pPlot:GetX(), pPlot:GetY(), pTestPlot:GetX(), pTestPlot:GetY() );
		
		-- if it's too close, return true
		if iDistanceFromTestPlot < iMinDistance then
			return true;
		end
	end 
	-- if none of the plots were too close, return false.
	return false;
end

-- accepts a plot and a feature type.
-- returns true if HV_Maps tables say this feature is permissable in this region
function HV_ValidFeatureRegion (sRegion, sFeatureType)
	
	if sRegion == nil or sFeatureType == nil then
		print("Error. Bad arguments to HV_ValidFeatureRegion");
	end

	--print(" - Checking suitablility of " .. sFeatureType .. " for region " .. sRegion);

	local aPermittedRegions = {};
	local iNumPermittedRegions = 0;
	
	-- check wonder region list
	for Row in GameInfo.HV_Map_TrueWonderRegion() do
		-- is row for this map type/size?
		if (Row.MapType == g_sMapType or Row.MapType == nil) and (Row.MapSize == g_sMapSize or Row.MapSize == nil) then
			-- is row for this wonder?
			if Row.Wonder == sFeatureType and Row.Region ~= nil then
				-- add this region to permitted  list for this wonder
				table.insert(aPermittedRegions, Row.Region);
				iNumPermittedRegions = iNumPermittedRegions +1;
				--print(" - " .. sFeatureType .. " permitted in region " .. Row.Region);
			end
		end
	end
	
	-- TO DO: make tables to exclude/permit generic features per Region, e.g. no jungle in England.
	
	-- if there are no entries about this region in the HV_Maps table, assume it's valid everywhere.
	--[[ or maybe assume it's valid nowhere, havn't decided yet.
	if iNumPermittedRegions == 0 then
		return true;
	end
	--]]
	
	-- if there are entries for this wonder in HV_Maps table, then it's only valid in one of those regions.
	if	TestMembership(aPermittedRegions, sRegion) == true then 
		--print("- Region is Valid!");
		return true;
	else
		return false;
	end
end

-- accepts a feature, returns list of regions HV_Maps tables say this feature is valid in 
function HV_FeatureValidRegions (sFeatureType)

	local aPermittedRegions = {};
	local iNumPermittedRegions = 0;
	
	-- check wonder region list
	for Row in GameInfo.HV_Map_TrueWonderRegion() do
		-- is row for this map type/size?
		if (Row.MapType == g_sMapType or Row.MapType == nil) and (Row.MapSize == g_sMapSize or Row.MapSize == nil) then
			-- is row for this wonder?
			if Row.Wonder == sFeatureType and Row.Region ~= nil then
				-- add this region to permitted  list for this wonder
				table.insert(aPermittedRegions, Row.Region);
				iNumPermittedRegions = iNumPermittedRegions +1;
				--print(" - " .. sFeatureType .. " permitted in region " .. Row.Region);
			end
		end
	end
	

	return aPermittedRegions;
	
end